//
//  AppDelegate.swift
//  Stepper
//
//  Created by Phidgets on 2018-02-21.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate, NSTabViewDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = Stepper()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var statusBox: NSBox!
    @IBOutlet weak var tabView: NSTabView!
    @IBOutlet weak var rescaleFactorText: NSTextField!
    @IBOutlet weak var motorPositionLabel: NSTextField!
    @IBOutlet weak var motorVelocityLabel: NSTextField!
    @IBOutlet weak var isMovingLabel: NSTextField!
    @IBOutlet weak var engageButton: NSButton!
        //position tab
    @IBOutlet weak var position_accelerationSlider: NSSlider!
    @IBOutlet weak var position_minAccelerationLabel: NSTextField!
    @IBOutlet weak var position_maxAccelerationLabel: NSTextField!
    @IBOutlet weak var position_accelerationLabel: NSTextField!
    @IBOutlet weak var position_velocityLimitSlider: NSSlider!
    @IBOutlet weak var position_minVelocityLimitLabel: NSTextField!
    @IBOutlet weak var position_maxVelocityLimitLabel: NSTextField!
    @IBOutlet weak var position_velocityLimitLabel: NSTextField!
    @IBOutlet weak var position_targetPositionSlider: NSSlider!
    @IBOutlet weak var position_minTargetPositionLabel: NSTextField!
    @IBOutlet weak var position_maxTargetPositionLabel: NSTextField!
    @IBOutlet weak var position_targetPositionLabel: NSTextField!
    @IBOutlet weak var position_currentLimitSlider: NSSlider!
    @IBOutlet weak var position_minCurrentLimitLabel: NSTextField!
    @IBOutlet weak var position_maxCurrentLimitLabel: NSTextField!
    @IBOutlet weak var position_currentLimitLabel: NSTextField!
    @IBOutlet weak var position_currentLimitBox: NSBox!
    @IBOutlet weak var position_holdingCurrentLimitBox: NSBox!
    @IBOutlet weak var position_targetPositionRangeText: NSTextField!
    @IBOutlet weak var position_holdingCurrentLimitText: NSTextField!
    
        //velocity tab
    @IBOutlet weak var velocity_accelerationSlider: NSSlider!
    @IBOutlet weak var velocity_minAccelerationLabel: NSTextField!
    @IBOutlet weak var velocity_maxAccelerationLabel: NSTextField!
    @IBOutlet weak var velocity_accelerationLabel: NSTextField!
    @IBOutlet weak var velocity_velocityLimitSlider: NSSlider!
    @IBOutlet weak var velocity_minVelocityLimitLabel: NSTextField!
    @IBOutlet weak var velocity_maxVelocityLimitLabel: NSTextField!
    @IBOutlet weak var velocity_velocityLimitLabel: NSTextField!
    @IBOutlet weak var velocity_currentLimitSlider: NSSlider!
    @IBOutlet weak var velocity_minCurrentLimitLabel: NSTextField!
    @IBOutlet weak var velocity_maxCurrentLimitLabel: NSTextField!
    @IBOutlet weak var velocity_currentLimitLabel: NSTextField!
    @IBOutlet weak var velocity_currentLimitBox: NSBox!
    @IBOutlet weak var velocity_holdingCurrentLimitBox: NSBox!
    @IBOutlet weak var velocity_holdingCurrentLimitText: NSTextField!

    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0
    
    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.positionChange.addHandler(positionchange_handler)
            let _ = ch.velocityChange.addHandler(velocitychange_handler)
            let _ = ch.stopped.addHandler(stopped_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.positionChange.removeAllHandlers()
            ch.velocityChange.removeAllHandlers()
            ch.stopped.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! Stepper
            let deviceID = try attachedDevice.getDeviceID()
            let minCurrentLimit = try attachedDevice.getMinCurrentLimit()
            let maxCurrentLimit = try attachedDevice.getMaxCurrentLimit()
            let currentLimit = try attachedDevice.getCurrentLimit()
            let engaged = try attachedDevice.getEngaged()
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.statusBox.isHidden = false
                self.tabView.isHidden = false
                
                self.updateGUI(isInitialUpdate: true)
                
                if(deviceID == DeviceID.PN_1062){
                    self.position_currentLimitBox.isHidden = true
                    self.velocity_currentLimitBox.isHidden = true
                }
                else{
                    self.position_currentLimitBox.isHidden = false
                    self.velocity_currentLimitBox.isHidden = false
                    
                    self.position_currentLimitSlider.minValue = minCurrentLimit
                    self.position_currentLimitSlider.maxValue = maxCurrentLimit
                    self.position_currentLimitSlider.doubleValue = currentLimit
                    self.position_minCurrentLimitLabel.stringValue = String(format:"%.2f", minCurrentLimit)
                    self.position_maxCurrentLimitLabel.stringValue = String(format:"%.2f", maxCurrentLimit)
                    self.position_currentLimitLabel.stringValue = String(format:"%.2f A", currentLimit)
                    
                    self.velocity_currentLimitSlider.minValue = minCurrentLimit
                    self.velocity_currentLimitSlider.maxValue = maxCurrentLimit
                    self.velocity_currentLimitSlider.doubleValue = currentLimit
                    self.velocity_minCurrentLimitLabel.stringValue = String(format:"%.2f", minCurrentLimit)
                    self.velocity_maxCurrentLimitLabel.stringValue = String(format:"%.2f", maxCurrentLimit)
                    self.velocity_currentLimitLabel.stringValue = String(format:"%.2f A", currentLimit)
                }
                
                if(deviceID == DeviceID.PN_STC1000){
                    self.position_holdingCurrentLimitBox.isHidden = false
                    self.velocity_holdingCurrentLimitBox.isHidden = false
                }
                else{
                    self.position_holdingCurrentLimitBox.isHidden = true
                    self.velocity_holdingCurrentLimitBox.isHidden = true
                }
                
                self.engageButton.title = engaged ? "Disengage" : "Engage"
                
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.statusBox.isHidden = true
            self.tabView.isHidden = true
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func positionchange_handler(sender: Stepper, position: Double){
        DispatchQueue.main.async {
            self.motorPositionLabel.stringValue = String(format:"%.3f",position)
            self.updateStopped()
        }
    }
    
    func velocitychange_handler(sender: Stepper, velocity: Double){
        DispatchQueue.main.async {
            self.motorVelocityLabel.stringValue = String(format:"%.3f",velocity)
            self.updateStopped()
        }
    }
    
    func stopped_handler(sender: Stepper){
        DispatchQueue.main.async {
            self.isMovingLabel.stringValue = "Stopped"
            self.isMovingLabel.backgroundColor = NSColor(calibratedRed: 1, green: 0.494, blue: 0.475, alpha: 1)
        }
    }
    
    //MARK: Tab View Delegate
    func tabView(_ tabView: NSTabView, didSelect tabViewItem: NSTabViewItem?) {
        do{
            if(tabView.indexOfTabViewItem(tabViewItem!) == 0){
                try ch.setControlMode(StepperControlMode.step)
            }
            else{
                try ch.setControlMode(StepperControlMode.run)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    //MARK: GUI Controls
    @IBAction func setTargetPositionRange(_ sender: Any) {
        do{
            if var range = Double(position_targetPositionRangeText.stringValue){
                range = fabs(range)
                let maxPosition = try ch.getMaxPosition()
                if(range > maxPosition){
                    position_targetPositionRangeText.stringValue = "Invalid"
                }
                else{
                    position_targetPositionSlider.minValue = -range
                    position_targetPositionSlider.maxValue = range
                    position_minTargetPositionLabel.stringValue = String(format:"%.3f", -range)
                    position_maxTargetPositionLabel.stringValue = String(format:"%.3f", range)
                }
            }
            else{
                position_targetPositionRangeText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setRescaleFactor(_ sender: Any) {
        do{
            if let rescaleFactor = Double(rescaleFactorText.stringValue){
                try ch.setRescaleFactor(rescaleFactor)
                self.updateGUI(isInitialUpdate: false)
            }
            else{
                let rescaleFactor = try ch.getRescaleFactor()
                rescaleFactorText.stringValue = String(format:"%.3e",rescaleFactor)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func engageMotor(_ sender: Any) {
        do{
            let button = sender as! NSButton
            let engaged = try ch.getEngaged()
            try ch.setEngaged(!engaged)
            if(engaged){
                button.title = "Engage"
            }
            else{
                button.title = "Disengage"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
        
    }
    
    @IBAction func setAcceleration(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setAcceleration(slider.doubleValue)
            position_accelerationLabel.stringValue = String(format:"%.3f", slider.doubleValue)
            velocity_accelerationLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setVelocityLimit(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setVelocityLimit(slider.doubleValue)
            position_velocityLimitLabel.stringValue = String(format:"%.3f", slider.doubleValue)
            velocity_velocityLimitLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCurrentLimit(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setCurrentLimit(slider.doubleValue)
            position_currentLimitLabel.stringValue = String(format:"%.2f A", slider.doubleValue)
            velocity_currentLimitLabel.stringValue = String(format:"%.2f A", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setTargetPosition(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetPosition(slider.doubleValue)
            position_targetPositionLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func zeroPosition(_ sender: Any) {
        do{
            var position = try ch.getPosition()
            try ch.addPositionOffset(positionOffset: -1 * position)
            position = try ch.getPosition()
            motorPositionLabel.stringValue = String(format:"%.3f", position)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setHoldingCurrent(_ sender: Any) {
        do{
            let button = sender as! NSButton
            let textField:NSTextField = button.tag == 0 ? position_holdingCurrentLimitText : velocity_holdingCurrentLimitText
            if let newHoldingCurrent = Double(textField.stringValue){
                try ch.setHoldingCurrentLimit(newHoldingCurrent)
            }
            else{
                textField.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Misc
    func updateStopped(){
        do{
            let engaged = try ch.getEngaged()
            if(!engaged){
                return
            }
            let isMoving = try ch.getIsMoving()
            DispatchQueue.main.async {
                if(isMoving){
                    self.isMovingLabel.stringValue = "Moving"
                    self.isMovingLabel.backgroundColor = NSColor(calibratedRed: 0, green: 0.561, blue: 0, alpha: 1)
                }
                else{
                    self.isMovingLabel.stringValue = "Stopped"
                    self.isMovingLabel.backgroundColor = NSColor(calibratedRed: 1, green: 0.494, blue: 0.475, alpha: 1)
                }
            }
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func updateGUI(isInitialUpdate:Bool){
        do{
            let controlMode = try ch.getControlMode()
            let rescaleFactor = try ch.getRescaleFactor()
            let minVelocityLimit = try ch.getMinVelocityLimit()
            let maxVelocityLimit = try ch.getMaxVelocityLimit()
            let velocityLimit = try ch.getVelocityLimit()
            let minAcceleration = try ch.getMinAcceleration()
            let maxAcceleration = try ch.getMaxAcceleration()
            let acceleration = try ch.getAcceleration()
            let targetPosition = try ch.getTargetPosition()
            let position = try ch.getPosition()
            let velocity = try ch.getVelocity()
            var minPosition = isInitialUpdate ? -10000 : try ch.getMinPosition()
            var maxPosition = isInitialUpdate ? 10000 : try ch.getMaxPosition()
            
            
            
            if(controlMode == StepperControlMode.step){
                minPosition = -10000*rescaleFactor
                maxPosition = 10000*rescaleFactor
                self.position_targetPositionRangeText.stringValue = String(format:"%.3f",10000*rescaleFactor)
            }
            
            if(isInitialUpdate){
                self.position_targetPositionRangeText.stringValue = "100000"
            }
            
            self.motorVelocityLabel.stringValue = String(format:"%.3f", velocity)
            self.motorPositionLabel.stringValue = String(format:"%.3f", position)
            self.position_velocityLimitSlider.minValue = minVelocityLimit
            self.position_velocityLimitSlider.maxValue = maxVelocityLimit
            self.position_velocityLimitSlider.doubleValue = velocityLimit
            self.position_minVelocityLimitLabel.stringValue = String(format:"%.3f", minVelocityLimit)
            self.position_maxVelocityLimitLabel.stringValue = String(format:"%.3f", maxVelocityLimit)
            self.position_velocityLimitLabel.stringValue = String(format:"%.3f",velocityLimit)
            
            self.position_accelerationSlider.minValue = minAcceleration
            self.position_accelerationSlider.maxValue = maxAcceleration
            self.position_accelerationSlider.doubleValue = acceleration
            self.position_minAccelerationLabel.stringValue = String(format:"%.3f", minAcceleration)
            self.position_maxAccelerationLabel.stringValue = String(format:"%.3f", maxAcceleration)
            self.position_accelerationLabel.stringValue = String(format:"%.3f", acceleration)
            
            self.position_targetPositionSlider.minValue = minPosition
            self.position_targetPositionSlider.maxValue = maxPosition
            self.position_targetPositionSlider.doubleValue = targetPosition
            self.position_minTargetPositionLabel.stringValue = String(format:"%.3f", minPosition)
            self.position_maxTargetPositionLabel.stringValue = String(format:"%.3f", maxPosition)
            self.position_targetPositionLabel.stringValue = String(format:"%.3f", targetPosition)
            
            
            self.velocity_velocityLimitSlider.minValue = minVelocityLimit
            self.velocity_velocityLimitSlider.maxValue = maxVelocityLimit
            self.velocity_velocityLimitSlider.doubleValue = velocityLimit
            self.velocity_minVelocityLimitLabel.stringValue = String(format:"%.3f", minVelocityLimit)
            self.velocity_maxVelocityLimitLabel.stringValue = String(format:"%.3f", maxVelocityLimit)
            self.velocity_velocityLimitLabel.stringValue = String(format:"%.3f", velocityLimit)
            
            self.velocity_accelerationSlider.minValue = minAcceleration
            self.velocity_accelerationSlider.maxValue = maxAcceleration
            self.velocity_accelerationSlider.doubleValue = acceleration
            self.velocity_minAccelerationLabel.stringValue = String(format:"%.3f", minAcceleration)
            self.velocity_maxAccelerationLabel.stringValue = String(format:"%.3f", maxAcceleration)
            self.velocity_accelerationLabel.stringValue = String(format:"%.3f", acceleration)
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
    
    
}

